@@ -7,6 +7,8 @@ pub use crate::error::DecodeError as Error;
 #[cfg(feature = "use_alloc")]
 use alloc::{ vec::Vec, string::String };
 
+#[cfg(feature = "use_alloc")]
+use crate::util::ScopeGuard;
 
 /// Read trait
 ///
@@ -375,20 +377,17 @@ fn decode_bytes<'a, R: Read<'a>>(name: &'static str, major_limit: u8, byte: u8,
 
 #[inline]
 #[cfg(feature = "use_alloc")]
-fn decode_buf<'a, R: Read<'a>>(major: u8, byte: u8, follow: bool, reader: &mut R, buf: &mut Vec<u8>)
+fn decode_buf<'a, R: Read<'a>>(major: u8, byte: u8, reader: &mut R, buf: &mut Vec<u8>)
     -> Result<Option<&'a [u8]>, Error<R::Error>>
 {
     const CAP_LIMIT: usize = 16 * 1024;
 
-    if follow && byte == marker::BREAK {
-        Ok(None)
-    } else if let Some(mut len) = decode_len(major, byte, reader)? {
-        if !follow {
-            if let Reference::Long(buf) = reader.fill(len)? {
-                if buf.len() >= len {
-                    reader.advance(len);
-                    return Ok(Some(&buf[..len]));
-                }
+    if let Some(mut len) = decode_len(major, byte, reader)? {
+        // try long lifetime buffer
+        if let Reference::Long(buf) = reader.fill(len)? {
+            if buf.len() >= len {
+                reader.advance(len);
+                return Ok(Some(&buf[..len]));
             }
         }
 
@@ -399,21 +398,40 @@ fn decode_buf<'a, R: Read<'a>>(major: u8, byte: u8, follow: bool, reader: &mut R
         while len != 0 {
             let readbuf = reader.fill(len)?;
             let readbuf = readbuf.as_ref();
-            let readlen = readbuf.len();
 
-            if readlen == 0 {
+            if readbuf.is_empty() {
                 return Err(Error::Eof);
             }
 
-            buf.extend_from_slice(readbuf);
+            let readlen = core::cmp::min(readbuf.len(), len);
+
+            buf.extend_from_slice(&readbuf[..readlen]);
             reader.advance(readlen);
             len -= readlen;
         }
 
         Ok(None)
     } else {
-        let byte = pull_one(reader)?;
-        decode_buf(major, byte, true, reader, buf)
+        // bytes sequence
+        loop {
+            let byte = pull_one(reader)?;
+
+            if byte == marker::BREAK {
+                break
+            }
+
+            if !reader.step_in() {
+                return Err(Error::DepthLimit);
+            }
+            let mut reader = ScopeGuard(reader, |reader| reader.step_out());
+            let reader = &mut *reader;
+
+            if let Some(longbuf) = decode_buf(major, byte, reader, buf)? {
+                buf.extend_from_slice(longbuf);
+            }
+        }
+
+        Ok(None)
     }
 }
 
@@ -430,7 +448,7 @@ impl<'a> Decode<'a> for types::Bytes<Vec<u8>> {
     #[inline]
     fn decode_with<R: Read<'a>>(byte: u8, reader: &mut R) -> Result<Self, Error<R::Error>> {
         let mut buf = Vec::new();
-        if let Some(longbuf) = decode_buf(major::BYTES, byte, false, reader, &mut buf)? {
+        if let Some(longbuf) = decode_buf(major::BYTES, byte, reader, &mut buf)? {
             buf.extend_from_slice(longbuf);
         }
         Ok(types::Bytes(buf))
@@ -444,7 +462,7 @@ impl<'a> Decode<'a> for types::Bytes<alloc::borrow::Cow<'a, [u8]>> {
         use alloc::borrow::Cow;
 
         let mut buf = Vec::new();
-        Ok(types::Bytes(if let Some(longbuf) = decode_buf(major::BYTES, byte, false, reader, &mut buf)? {
+        Ok(types::Bytes(if let Some(longbuf) = decode_buf(major::BYTES, byte, reader, &mut buf)? {
             Cow::Borrowed(longbuf)
         } else {
             Cow::Owned(buf)
@@ -466,7 +484,7 @@ impl<'a> Decode<'a> for String {
     #[inline]
     fn decode_with<R: Read<'a>>(byte: u8, reader: &mut R) -> Result<Self, Error<R::Error>> {
         let mut buf = Vec::new();
-        if let Some(longbuf) = decode_buf(major::STRING, byte, false, reader, &mut buf)? {
+        if let Some(longbuf) = decode_buf(major::STRING, byte, reader, &mut buf)? {
             buf.extend_from_slice(longbuf);
         }
         let buf = String::from_utf8(buf)
@@ -483,7 +501,7 @@ impl<'a> Decode<'a> for alloc::borrow::Cow<'a, str> {
         use alloc::borrow::Cow;
 
         let mut buf = Vec::new();
-        Ok(if let Some(longbuf) = decode_buf(major::STRING, byte, false, reader, &mut buf)? {
+        Ok(if let Some(longbuf) = decode_buf(major::STRING, byte, reader, &mut buf)? {
             Cow::Borrowed(core::str::from_utf8(longbuf).map_err(Error::InvalidUtf8)?)
         } else {
             let buf = String::from_utf8(buf)
@@ -506,7 +524,7 @@ impl<'a> Decode<'a> for types::BadStr<Vec<u8>> {
     #[inline]
     fn decode_with<R: Read<'a>>(byte: u8, reader: &mut R) -> Result<Self, Error<R::Error>> {
         let mut buf = Vec::new();
-        if let Some(longbuf) = decode_buf(major::STRING, byte, false, reader, &mut buf)? {
+        if let Some(longbuf) = decode_buf(major::STRING, byte, reader, &mut buf)? {
             buf.extend_from_slice(longbuf);
         }
         Ok(types::BadStr(buf))
@@ -520,7 +538,7 @@ impl<'a> Decode<'a> for types::BadStr<alloc::borrow::Cow<'a, [u8]>> {
         use alloc::borrow::Cow;
 
         let mut buf = Vec::new();
-        Ok(types::BadStr(if let Some(longbuf) = decode_buf(major::STRING, byte, false, reader, &mut buf)? {
+        Ok(types::BadStr(if let Some(longbuf) = decode_buf(major::STRING, byte, reader, &mut buf)? {
             Cow::Borrowed(longbuf)
         } else {
             Cow::Owned(buf)
@@ -548,6 +566,12 @@ impl<'a, T: Decode<'a>> Decode<'a> for Vec<T> {
     fn decode_with<R: Read<'a>>(byte: u8, reader: &mut R) -> Result<Self, Error<R::Error>> {
         let mut arr = Vec::new();
 
+        if !reader.step_in() {
+            return Err(Error::DepthLimit);
+        }
+        let mut reader = ScopeGuard(reader, |reader| reader.step_out());
+        let reader = &mut *reader;
+
         if let Some(len) = decode_len(major::ARRAY, byte, reader)? {
             if len <= 256 {
                 arr.reserve(len); // TODO try_reserve ?
@@ -580,6 +604,12 @@ impl<'a, K: Decode<'a>, V: Decode<'a>> Decode<'a> for types::Map<Vec<(K, V)>> {
     fn decode_with<R: Read<'a>>(byte: u8, reader: &mut R) -> Result<Self, Error<R::Error>> {
         let mut map = Vec::new();
 
+        if !reader.step_in() {
+            return Err(Error::DepthLimit);
+        }
+        let mut reader = ScopeGuard(reader, |reader| reader.step_out());
+        let reader = &mut *reader;
+
         if let Some(len) = decode_len(major::MAP, byte, reader)? {
             if len <= 256 {
                 map.reserve(len); // TODO try_reserve ?
