@@ -1,5 +1,4 @@
-use alloc::vec::Vec;
-use alloc::string::String;
+use alloc::borrow::Cow;
 use serde::de::{ self, Visitor };
 use crate::core::{ major, marker, types };
 use crate::core::dec::{ self, Decode };
@@ -20,8 +19,20 @@ impl<R> Deserializer<R> {
     }
 }
 
+impl<'de, R: dec::Read<'de>> Deserializer<R> {
+    #[inline]
+    fn try_step(&mut self) -> Result<ScopeGuard<'_, Self>, dec::Error<R::Error>> {
+        if self.reader.step_in() {
+            Ok(ScopeGuard(self, |de| de.reader.step_out()))
+        } else {
+            Err(dec::Error::DepthLimit)
+        }
+    }
+}
+
 macro_rules! deserialize_type {
     ( @ $t:ty , $name:ident , $visit:ident ) => {
+        #[inline]
         fn $name<V>(self, visitor: V) -> Result<V::Value, Self::Error>
         where V: Visitor<'de>
         {
@@ -39,15 +50,13 @@ macro_rules! deserialize_type {
 impl<'de, 'a, R: dec::Read<'de>> serde::Deserializer<'de> for &'a mut Deserializer<R> {
     type Error = dec::Error<R::Error>;
 
-    fn deserialize_any<V>(mut self, visitor: V) -> Result<V::Value, Self::Error>
+    #[inline]
+    fn deserialize_any<V>(self, visitor: V) -> Result<V::Value, Self::Error>
     where
         V: Visitor<'de>
     {
-        if !self.reader.step_in() {
-            return Err(dec::Error::RecursionLimit);
-        }
-        let mut de = ScopeGuard(&mut self, |de| de.reader.step_out());
-        let de = de.get_mut();
+        let mut de = self.try_step()?;
+        let de = &mut *de;
 
         let byte = dec::peek_one(&mut de.reader)?;
         match byte >> 5 {
@@ -57,7 +66,13 @@ impl<'de, 'a, R: dec::Read<'de>> serde::Deserializer<'de> for &'a mut Deserializ
             major::STRING => de.deserialize_string(visitor),
             major::ARRAY => de.deserialize_seq(visitor),
             major::MAP => de.deserialize_map(visitor),
-            _ => match byte {
+            // TODO serde support https://github.com/serde-rs/serde/issues/1682
+            // major::TAG => match byte {
+            //     marker::BIGNUM => de.deserialize_u128(visitor),
+            //     marker::NEG_BIGNUM => de.deserialize_i128(visitor),
+            //     _ => Err(dec::Error::Unsupported { byte })
+            // },
+            major::SIMPLE => match byte {
                 marker::FALSE => {
                     de.reader.advance(1);
                     visitor.visit_bool(false)
@@ -79,7 +94,8 @@ impl<'de, 'a, R: dec::Read<'de>> serde::Deserializer<'de> for &'a mut Deserializ
                 marker::F32 => de.deserialize_f32(visitor),
                 marker::F64 => de.deserialize_f32(visitor),
                 _ => Err(dec::Error::Unsupported { byte })
-            }
+            },
+            _ => Err(dec::Error::Unsupported { byte })
         }
     }
 
@@ -100,11 +116,9 @@ impl<'de, 'a, R: dec::Read<'de>> serde::Deserializer<'de> for &'a mut Deserializ
 
         f32,        deserialize_f32,        visit_f32;
         f64,        deserialize_f64,        visit_f64;
-
-        &str,       deserialize_str,        visit_borrowed_str;
-        String,     deserialize_string,     visit_string;
     );
 
+    #[inline]
     fn deserialize_char<V>(self, visitor: V) -> Result<V::Value, Self::Error>
     where V: Visitor<'de>
     {
@@ -128,33 +142,57 @@ impl<'de, 'a, R: dec::Read<'de>> serde::Deserializer<'de> for &'a mut Deserializ
         }
     }
 
+    #[inline]
     fn deserialize_bytes<V>(self, visitor: V) -> Result<V::Value, Self::Error>
     where V: Visitor<'de>
     {
-        let value = <types::Bytes<&[u8]>>::decode(&mut self.reader)?;
-        visitor.visit_borrowed_bytes(value.0)
+        let v = <types::Bytes<&[u8]>>::decode(&mut self.reader)?;
+        visitor.visit_borrowed_bytes(v.0)
     }
 
-
+    #[inline]
     fn deserialize_byte_buf<V>(self, visitor: V) -> Result<V::Value, Self::Error>
     where V: Visitor<'de>
     {
-        let value = <types::Bytes<Vec<u8>>>::decode(&mut self.reader)?;
-        visitor.visit_byte_buf(value.0)
+        match <types::Bytes<Cow<[u8]>>>::decode(&mut self.reader)?.0 {
+            Cow::Borrowed(buf) => visitor.visit_borrowed_bytes(buf),
+            Cow::Owned(buf) => visitor.visit_byte_buf(buf)
+        }
     }
 
+    #[inline]
+    fn deserialize_str<V>(self, visitor: V) -> Result<V::Value, Self::Error>
+    where V: Visitor<'de>
+    {
+        let v = <&str>::decode(&mut self.reader)?;
+        visitor.visit_borrowed_str(v)
+    }
+
+    #[inline]
+    fn deserialize_string<V>(self, visitor: V) -> Result<V::Value, Self::Error>
+    where V: Visitor<'de>
+    {
+        match <Cow<str>>::decode(&mut self.reader)? {
+            Cow::Borrowed(buf) => visitor.visit_borrowed_str(buf),
+            Cow::Owned(buf) => visitor.visit_string(buf)
+        }
+    }
+
+    #[inline]
     fn deserialize_option<V>(self, visitor: V) -> Result<V::Value, Self::Error>
     where V: Visitor<'de>
     {
         let byte = dec::peek_one(&mut self.reader)?;
         if byte != marker::NULL && byte != marker::UNDEFINED {
-            visitor.visit_some(self)
+            let mut de = self.try_step()?;
+            visitor.visit_some(&mut *de)
         } else {
             self.reader.advance(1);
             visitor.visit_none()
         }
     }
 
+    #[inline]
     fn deserialize_unit<V>(self, visitor: V) -> Result<V::Value, Self::Error>
     where V: Visitor<'de>
     {
@@ -169,6 +207,7 @@ impl<'de, 'a, R: dec::Read<'de>> serde::Deserializer<'de> for &'a mut Deserializ
         }
     }
 
+    #[inline]
     fn deserialize_unit_struct<V>(self, _name: &'static str, visitor: V)
         -> Result<V::Value, Self::Error>
     where V: Visitor<'de>
@@ -176,6 +215,7 @@ impl<'de, 'a, R: dec::Read<'de>> serde::Deserializer<'de> for &'a mut Deserializ
         self.deserialize_unit(visitor)
     }
 
+    #[inline]
     fn deserialize_newtype_struct<V>(self, _name: &'static str, visitor: V)
         -> Result<V::Value, Self::Error>
     where V: Visitor<'de>
@@ -183,13 +223,16 @@ impl<'de, 'a, R: dec::Read<'de>> serde::Deserializer<'de> for &'a mut Deserializ
         visitor.visit_newtype_struct(self)
     }
 
+    #[inline]
     fn deserialize_seq<V>(self, visitor: V) -> Result<V::Value, Self::Error>
     where V: Visitor<'de>
     {
-        let seq = Accessor::array(self)?;
+        let mut de = self.try_step()?;
+        let seq = Accessor::array(&mut de)?;
         visitor.visit_seq(seq)
     }
 
+    #[inline]
     fn deserialize_tuple<V>(
         self,
         len: usize,
@@ -198,10 +241,12 @@ impl<'de, 'a, R: dec::Read<'de>> serde::Deserializer<'de> for &'a mut Deserializ
     where
         V: Visitor<'de>
     {
-        let seq = Accessor::tuple(self, len)?;
+        let mut de = self.try_step()?;
+        let seq = Accessor::tuple(&mut de, len)?;
         visitor.visit_seq(seq)
     }
 
+    #[inline]
     fn deserialize_tuple_struct<V>(
         self,
         _name: &'static str,
@@ -214,14 +259,17 @@ impl<'de, 'a, R: dec::Read<'de>> serde::Deserializer<'de> for &'a mut Deserializ
         self.deserialize_tuple(len, visitor)
     }
 
+    #[inline]
     fn deserialize_map<V>(self, visitor: V) -> Result<V::Value, Self::Error>
     where
         V: Visitor<'de>
     {
-        let map = Accessor::map(self)?;
+        let mut de = self.try_step()?;
+        let map = Accessor::map(&mut de)?;
         visitor.visit_map(map)
     }
 
+    #[inline]
     fn deserialize_struct<V>(
         self,
         _name: &'static str,
@@ -233,6 +281,7 @@ impl<'de, 'a, R: dec::Read<'de>> serde::Deserializer<'de> for &'a mut Deserializ
         self.deserialize_map(visitor)
     }
 
+    #[inline]
     fn deserialize_enum<V>(
         self,
         _name: &'static str,
@@ -242,31 +291,12 @@ impl<'de, 'a, R: dec::Read<'de>> serde::Deserializer<'de> for &'a mut Deserializ
     where
         V: Visitor<'de>
     {
-        let byte = dec::peek_one(&mut self.reader)?;
-        let accessor = match byte >> 5 {
-            major::STRING => EnumAccessor { de: self },
-            major::MAP => {
-                self.reader.advance(1);
-                let len = dec::decode_len(major::MAP, byte, &mut self.reader)?;
-                if len != Some(1) {
-                    return Err(dec::Error::RequireLength {
-                        name: "enum::map",
-                        expect: 1,
-                        value: len.unwrap_or(0)
-                    });
-                }
-
-                EnumAccessor { de: self }
-            },
-            _ => return Err(dec::Error::TypeMismatch {
-                name: "enum",
-                byte
-            })
-        };
-
+        let mut de = self.try_step()?;
+        let accessor = EnumAccessor::enum_(&mut de)?;
         visitor.visit_enum(accessor)
     }
 
+    #[inline]
     fn deserialize_identifier<V>(self, visitor: V)
         -> Result<V::Value, Self::Error>
     where V: Visitor<'de>
@@ -274,14 +304,66 @@ impl<'de, 'a, R: dec::Read<'de>> serde::Deserializer<'de> for &'a mut Deserializ
         self.deserialize_str(visitor)
     }
 
+    #[inline]
     fn deserialize_ignored_any<V>(self, visitor: V)
         -> Result<V::Value, Self::Error>
     where V: Visitor<'de>
     {
-        // TODO skip length optimize
-        self.deserialize_any(visitor)
+        let mut de = self.try_step()?;
+        let byte = dec::peek_one(&mut de.reader)?;
+
+        match byte >> 5 {
+            major @ major::UNSIGNED | major @ major::NEGATIVE => {
+                let skip = match byte & !(major << 5) {
+                    0 ..= 0x17 => 1,
+                    0x18 => 2,
+                    0x19 => 3,
+                    0x1a => 5,
+                    0x1b => 9,
+                    _ => return Err(dec::Error::TypeMismatch {
+                        name: "any",
+                        byte
+                    })
+                };
+                skip_exact(&mut de.reader, skip)?;
+            },
+            major @ major::BYTES | major @ major::STRING |
+            major @ major::ARRAY | major @ major::MAP => {
+                de.reader.advance(1);
+
+                if let Some(len) = dec::decode_len(major, byte, &mut de.reader)? {
+                    match major {
+                        major::BYTES | major::STRING => skip_exact(&mut de.reader, len)?,
+                        major::ARRAY | major::MAP => for _ in 0..len {
+                            de.deserialize_ignored_any(de::IgnoredAny)?;
+
+                            if major == major::MAP {
+                                de.deserialize_ignored_any(de::IgnoredAny)?;
+                            }
+                        },
+                        _ => ()
+                    }
+                } else {
+                    while dec::peek_one(&mut de.reader)? != marker::BREAK {
+                        de.deserialize_ignored_any(de::IgnoredAny)?;
+
+                        if major == major::MAP {
+                            de.deserialize_ignored_any(de::IgnoredAny)?;
+                        }
+                    }
+                }
+            },
+            major @ major::TAG => {
+                let _tag = dec::TypeNum::new(!(major << 5), byte).decode_u8(&mut de.reader)?;
+                de.deserialize_ignored_any(de::IgnoredAny)?;
+            },
+            _ => return de.deserialize_any(visitor)
+        }
+
+        visitor.visit_unit()
     }
 
+    #[inline]
     fn is_human_readable(&self) -> bool {
         false
     }
@@ -293,6 +375,7 @@ struct Accessor<'a, R> {
 }
 
 impl<'de, 'a, R: dec::Read<'de>> Accessor<'a, R> {
+    #[inline]
     pub fn array(de: &'a mut Deserializer<R>)
         -> Result<Accessor<'a, R>, dec::Error<R::Error>>
     {
@@ -301,6 +384,7 @@ impl<'de, 'a, R: dec::Read<'de>> Accessor<'a, R> {
         Ok(Accessor { de, len })
     }
 
+    #[inline]
     pub fn tuple(de: &'a mut Deserializer<R>, len: usize)
         -> Result<Accessor<'a, R>, dec::Error<R::Error>>
     {
@@ -318,6 +402,7 @@ impl<'de, 'a, R: dec::Read<'de>> Accessor<'a, R> {
         }
     }
 
+    #[inline]
     pub fn map(de: &'a mut Deserializer<R>)
         -> Result<Accessor<'a, R>, dec::Error<R::Error>>
     {
@@ -333,6 +418,7 @@ where
 {
     type Error = dec::Error<R::Error>;
 
+    #[inline]
     fn next_element_seed<T>(&mut self, seed: T)
         -> Result<Option<T::Value>, Self::Error>
     where T: de::DeserializeSeed<'de>
@@ -352,6 +438,7 @@ where
         }
     }
 
+    #[inline]
     fn size_hint(&self) -> Option<usize> {
         self.len
     }
@@ -360,6 +447,7 @@ where
 impl<'de, 'a, R: dec::Read<'de>> de::MapAccess<'de> for Accessor<'a, R> {
     type Error = dec::Error<R::Error>;
 
+    #[inline]
     fn next_key_seed<K>(&mut self, seed: K) -> Result<Option<K::Value>, Self::Error>
     where K: de::DeserializeSeed<'de>
     {
@@ -378,12 +466,14 @@ impl<'de, 'a, R: dec::Read<'de>> de::MapAccess<'de> for Accessor<'a, R> {
         }
     }
 
+    #[inline]
     fn next_value_seed<V>(&mut self, seed: V) -> Result<V::Value, Self::Error>
     where V: de::DeserializeSeed<'de>
     {
         seed.deserialize(&mut *self.de)
     }
 
+    #[inline]
     fn size_hint(&self) -> Option<usize> {
         self.len
     }
@@ -393,6 +483,35 @@ struct EnumAccessor<'a, R> {
     de: &'a mut Deserializer<R>,
 }
 
+impl<'de, 'a, R: dec::Read<'de>> EnumAccessor<'a, R> {
+    #[inline]
+    pub fn enum_(de: &'a mut Deserializer<R>)
+        -> Result<EnumAccessor<'a, R>, dec::Error<R::Error>>
+    {
+        let byte = dec::peek_one(&mut de.reader)?;
+        match byte >> 5 {
+            major::STRING => Ok(EnumAccessor { de }),
+            major::MAP => {
+                de.reader.advance(1);
+
+                // 1 length map
+                if byte == (major::MAP << 5) | 1 {
+                    Ok(EnumAccessor { de })
+                } else {
+                    Err(dec::Error::TypeMismatch {
+                        name: "enum::map",
+                        byte
+                    })
+                }
+            },
+            _ => Err(dec::Error::TypeMismatch {
+                name: "enum",
+                byte
+            })
+        }
+    }
+}
+
 impl<'de, 'a, R> de::EnumAccess<'de> for EnumAccessor<'a, R>
 where
     R: dec::Read<'de>
@@ -400,6 +519,7 @@ where
     type Error = dec::Error<R::Error>;
     type Variant = EnumAccessor<'a, R>;
 
+    #[inline]
     fn variant_seed<V>(self, seed: V)
         -> Result<(V::Value, Self::Variant), Self::Error>
     where V: de::DeserializeSeed<'de>
@@ -415,16 +535,19 @@ where
 {
     type Error = dec::Error<R::Error>;
 
+    #[inline]
     fn unit_variant(self) -> Result<(), Self::Error> {
         Ok(())
     }
 
+    #[inline]
     fn newtype_variant_seed<T>(self, seed: T) -> Result<T::Value, Self::Error>
     where T: de::DeserializeSeed<'de>
     {
         seed.deserialize(&mut *self.de)
     }
 
+    #[inline]
     fn tuple_variant<V>(self, len: usize, visitor: V)
         -> Result<V::Value, Self::Error>
     where V: Visitor<'de>
@@ -434,6 +557,7 @@ where
         self.de.deserialize_tuple(len, visitor)
     }
 
+    #[inline]
     fn struct_variant<V>(
         self,
         _fields: &'static [&'static str],
@@ -446,3 +570,17 @@ where
         self.de.deserialize_map(visitor)
     }
 }
+
+#[inline]
+fn skip_exact<'de, R: dec::Read<'de>>(reader: &mut R, mut len: usize) -> Result<(), R::Error> {
+    while len != 0 {
+        let buf = reader.fill(len)?;
+        let buf = buf.as_ref();
+
+        let buflen = core::cmp::min(len, buf.len());
+        reader.advance(buflen);
+        len -= buflen;
+    }
+
+    Ok(())
+}
